@@ -43,7 +43,7 @@ def _MySQL_concat_expr(dbmodel, expression):
 
 # map from op-name to special SQL formatting code
 MySQL_formatters = {
-    "___": lambda dbmodel, expression: expression.to_python(),
+    "___": lambda dbmodel, expression: str(expression.to_python()),
     "is_bad": _MySQL_is_bad_expr,
     "concat": _MySQL_concat_expr,
 }
@@ -51,10 +51,10 @@ MySQL_formatters = {
 
 class MySQLModel(data_algebra.db_model.DBModel):
     """A model of how SQL should be generated for MySQL.
-       Assuming we are using a sqlalhemy engine as our connection
+       Assuming we are using a sqlalchemy engine as our connection.
     """
 
-    def __init__(self):
+    def __init__(self, *, supports_with=False):
         data_algebra.db_model.DBModel.__init__(
             self,
             string_type="CHAR",
