@@ -55,7 +55,7 @@ def _sparksql_coalesce_expr(dbmodel, expression):
 
 # map from op-name to special SQL formatting code
 SparkSQL_formatters = {
-    "___": lambda dbmodel, expression: expression.to_python(),
+    "___": lambda dbmodel, expression: str(expression.to_python()),
     "is_bad": _sparksql_is_bad_expr,
     "coalesce": _sparksql_coalesce_expr,
 }
