@@ -1,5 +1,5 @@
 __docformat__ = "restructuredtext"
-__version__ = "0.8.0"
+__version__ = "0.9.0"
 
 __doc__ = """
 `data_algebra`<https://github.com/WinVector/data_algebra> is a piped data wrangling system
@@ -17,9 +17,7 @@ the `rquery`<https://github.com/WinVector/rquery> and `rqdatatable`<https://gith
 
 import data_algebra.pandas_model
 
-# set up what pandas supplier we are using
-import pandas
 
-default_data_model = data_algebra.pandas_model.PandasModel(
-    pd=pandas, presentation_model_name="pd"
-)
+# set up what pandas supplier we are using
+default_data_model = data_algebra.pandas_model.PandasModel()
+
