@@ -1,5 +1,4 @@
 import re
-import collections
 
 import numpy
 
@@ -8,6 +7,9 @@ import data_algebra.util
 
 
 class RecordSpecification:
+    """
+    Class to represent a multi-row data record.
+    """
     def __init__(
         self,
         control_table,
@@ -18,11 +20,11 @@ class RecordSpecification:
         local_data_model=None
     ):
         """
-        Class to represent a multi-row data record.
-
         :param control_table: data.frame describing record layout
         :param record_keys: array of record key column names
-        :param control_table_keys: array of control_table key column names
+               defaults to no columns.
+        :param control_table_keys: array of control_table key column names,
+               defaults to first column for non-trivial blocks and no columns for rows.
         :param strict: logical, if True more checks on transform
         :param local_data_model: data.frame data model
         """
@@ -33,24 +35,34 @@ class RecordSpecification:
             raise ValueError("control table should have at least 1 row")
         if len(control_table.columns) != len(set(control_table.columns)):
             raise ValueError("control table columns should be unique")
-        self.control_table = control_table.reset_index(drop=True)
+        self.control_table = control_table.reset_index(drop=True, inplace=False)
+        assert self.control_table.shape[0] > 0
         if record_keys is None:
             record_keys = []
         if isinstance(record_keys, str):
             record_keys = [record_keys]
         self.record_keys = [k for k in record_keys]
         if control_table_keys is None:
-            control_table_keys = [control_table.columns[0]]
-        if len(control_table_keys) <= 0:
-            raise ValueError("must have at least one control table key")
+            if self.control_table.shape[0] > 1:
+                control_table_keys = [self.control_table.columns[0]]
+            else:
+                control_table_keys = []  # single row records don't need to be keyed
         if isinstance(control_table_keys, str):
             control_table_keys = [control_table_keys]
+        if self.control_table.shape[0] == 1:
+            if len(control_table_keys) > 0:
+                raise ValueError("signle-row records must not have control table keys")
+        else:
+            if len(control_table_keys) <= 0:
+                raise ValueError("multi-row records must have at least one control table key")
         self.control_table_keys = [k for k in control_table_keys]
         unknown = set(self.control_table_keys) - set(control_table.columns)
         if len(unknown) > 0:
             raise ValueError(
                 "control table keys that are not in the control table: " + str(unknown)
             )
+        if len(self.control_table_keys) >= control_table.shape[1]:
+            raise ValueError("control table columns must not all be keys")
         confused = set(record_keys).intersection(control_table_keys)
         if len(confused) > 0:
             raise ValueError(
@@ -137,25 +149,23 @@ class RecordSpecification:
     def __str__(self):
         return self.fmt()
 
-    def map_to_rows(self, *, strict=False):
+    def map_to_rows(self):
         """
         Build a RecordMap mapping this RecordSpecification to rowrecs
 
-        :param strict:
         :return: RecordMap
         """
 
-        return RecordMap(blocks_in=self, strict=strict)
+        return RecordMap(blocks_in=self)
 
     def map_from_rows(self, *, strict=False):
         """
         Build a RecordMap mapping this RecordSpecification from rowrecs
 
-        :param strict:
         :return: RecordMap
         """
 
-        return RecordMap(blocks_out=self, strict=strict)
+        return RecordMap(blocks_out=self)
 
 
 def blocks_to_rowrecs(data, *, blocks_in, local_data_model=None):
@@ -227,6 +237,9 @@ def blocks_to_rowrecs(data, *, blocks_in, local_data_model=None):
     for c in blocks_in.row_version():
         if c not in colset:
             res[c] = None
+    if data.shape[0] <= 0:
+        res = res.loc[range(0), :]
+        res = res.reset_index(inplace=False, drop=True)
     return res
 
 
@@ -299,11 +312,14 @@ def rowrecs_to_blocks(
     if data.shape[0] < 1:
         # empty input produces emtpy output (with different column structure)
         res = res.iloc[range(0), :].reset_index(drop=True)
+    if data.shape[0] <= 0:
+        res = res.loc[range(0), :]
+        res = res.reset_index(inplace=False, drop=True)
     return res
 
 
 class RecordMap:
-    def __init__(self, *, blocks_in=None, blocks_out=None, strict=False):
+    def __init__(self, *, blocks_in=None, blocks_out=None):
         if blocks_in is not None:
             assert isinstance(blocks_in, data_algebra.cdata.RecordSpecification)
             ck = [k for k in blocks_in.content_keys if k is not None]
@@ -322,15 +338,6 @@ class RecordMap:
             unknown = set(blocks_out.content_keys) - set(blocks_in.content_keys)
             if len(unknown) > 0:
                 raise ValueError("unknown outgoing content_keys" + str(unknown))
-            if strict:
-                if set(blocks_in.record_keys) != set(blocks_out.record_keys):
-                    raise ValueError(
-                        "record keys must match when using both blocks in and blocks out"
-                    )
-                if set(blocks_in.content_keys) != set(blocks_out.content_keys):
-                    raise ValueError(
-                        "content keys must match when using both blocks in and blocks out"
-                    )
         self.blocks_in = blocks_in
         self.blocks_out = blocks_out
         if self.blocks_in is not None:
@@ -552,7 +559,6 @@ def pivot_blocks_to_rowrecs(
     attribute_value_column,
     record_keys,
     record_value_columns,
-    strict=False,
     local_data_model=None
 ):
     """
@@ -562,7 +568,6 @@ def pivot_blocks_to_rowrecs(
     :param attribute_value_column: column for record attribute values
     :param record_keys: names of key columns identifying row record blocks
     :param record_value_columns: names of columns to take row record values from
-    :param strict: logical, if True more checks on transform
     :param local_data_model: data.frame data model
     :return: RecordMap
     """
@@ -579,10 +584,9 @@ def pivot_blocks_to_rowrecs(
         control_table,
         record_keys=record_keys,
         control_table_keys=[attribute_key_column],
-        strict=strict,
         local_data_model=local_data_model,
     )
-    return ct.map_to_rows(strict=strict)
+    return ct.map_to_rows()
 
 
 def pivot_rowrecs_to_blocks(
@@ -591,7 +595,6 @@ def pivot_rowrecs_to_blocks(
     attribute_value_column,
     record_keys,
     record_value_columns,
-    strict=False,
     local_data_model=None
 ):
     """
@@ -601,7 +604,6 @@ def pivot_rowrecs_to_blocks(
     :param attribute_value_column: column for record attribute values
     :param record_keys: names of key columns identifying row record blocks
     :param record_value_columns: names of columns to take row record values from
-    :param strict: logical, if True more checks on transform
     :param local_data_model: data.frame data model
     :return: RecordMap
     """
@@ -618,7 +620,6 @@ def pivot_rowrecs_to_blocks(
         control_table,
         record_keys=record_keys,
         control_table_keys=[attribute_key_column],
-        strict=strict,
         local_data_model=local_data_model,
     )
-    return ct.map_from_rows(strict=strict)
+    return ct.map_from_rows()
