@@ -104,16 +104,14 @@ class ViewRepresentation(OperatorPlatform, ABC):
         :return: table result
         """
         tables = self.get_tables()
-        for t in tables.values():
-            assert isinstance(t, TableDescription)
-            assert t.head is not None
+        for tv in tables.values():
+            assert isinstance(tv, TableDescription)
+            assert tv.head is not None
             if len(tables) > 1:
-                assert t.table_name_was_set_by_user
+                assert tv.table_name_was_set_by_user
             if not allow_limited_tables:
-                assert t.limit_was is None
-                assert t.nrows == t.head.shape[0]
-        data_map = {k: v.head for k, v in tables.items()}
-        return self.eval(data_map=data_map, data_model=data_model, narrow=narrow)
+                assert tv.nrows == tv.head.shape[0]
+        return self.eval(data_map=None, data_model=data_model, narrow=narrow)
 
     # characterization
 
@@ -220,10 +218,10 @@ class ViewRepresentation(OperatorPlatform, ABC):
         return python_str
 
     def __repr__(self):
-        return self.to_python(strict=True)
+        return self.to_python(strict=True, pretty=True)
 
     def __str__(self):
-        return self.to_python(strict=True)
+        return self.to_python(strict=True, pretty=True)
 
     # noinspection PyMethodMayBeStatic,PyUnusedLocal
     def _equiv_nodes(self, other):
@@ -259,7 +257,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         raise NotImplementedError("base method called")
 
     def to_sql(
@@ -315,23 +313,23 @@ class ViewRepresentation(OperatorPlatform, ABC):
          :return: table result
          """
 
-        assert isinstance(data_map, dict)
-        if len(data_map) < 1:
-            raise ValueError("Expected data_map to be non-empty")
+        if data_map is not None:
+            assert isinstance(data_map, dict)
         if data_model is None:
             data_model = data_algebra.default_data_model
         assert isinstance(data_model, data_algebra.data_model.DataModel)
-        self.columns_used()  # for table consistency check/raise
-        tables = self.get_tables()
-        self.check_constraints(
-            {k: x.columns for (k, x) in data_map.items()}, strict=not narrow
-        )
-        for k in tables.keys():
-            if k not in data_map.keys():
-                raise ValueError("Required table " + k + " not in data_map")
-            else:
-                if not data_model.is_appropriate_data_instance(data_map[k]):
-                    raise ValueError("data_map[" + k + "] was not a usable type")
+        if (data_map is not None) and (len(data_map) > 0):
+            self.columns_used()  # for table consistency check/raise
+            tables = self.get_tables()
+            self.check_constraints(
+                {k: x.columns for (k, x) in data_map.items()}, strict=not narrow
+            )
+            for k in tables.keys():
+                if k not in data_map.keys():
+                    raise ValueError("Required table " + k + " not in data_map")
+                else:
+                    if not data_model.is_appropriate_data_instance(data_map[k]):
+                        raise ValueError("data_map[" + k + "] was not a usable type")
         return self.eval_implementation(
             data_map=data_map, data_model=data_model, narrow=narrow
         )
@@ -666,6 +664,8 @@ class TableDescription(ViewRepresentation):
     def same_table(self, other):
         if not isinstance(other, data_algebra.data_ops.TableDescription):
             return False
+        if self.table_name_was_set_by_user != other.table_name_was_set_by_user:
+            return False
         if self.table_name != other.table_name:
             return False
         if self.key != other.key:
@@ -761,7 +761,7 @@ class TableDescription(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.table_def_to_sql(
             self,
             using=using,
@@ -844,13 +844,13 @@ def describe_table(
     )
 
 
-def table(d, table_name=None):
+def table(d, *, table_name=None):
     """
     Capture a table for later use
 
     :param d: Pandas data frame to capture
     :param table_name: name for this table
-    :return: a table description
+    :return: a table description, with values retained
     """
     return describe_table(
         d=d,
@@ -864,6 +864,45 @@ def table(d, table_name=None):
     )
 
 
+def descr(**kwargs):
+    """
+    Capture a named partial table as a description.
+
+    :param kwargs: exactly one named table of the form table_name=table_value
+    :return: a table description (not all values retained)
+    """
+    assert len(kwargs) == 1
+    table_name = [k for k in kwargs.keys()][0]
+    d = kwargs[table_name]
+    return describe_table(
+        d=d,
+        table_name=table_name,
+        qualifiers=None,
+        sql_meta=None,
+        column_types=None,
+        row_limit=7,
+        keep_sample=True,
+        keep_all=False,
+    )
+
+
+def data(*args, **kwargs):
+    """
+    Capture a full table for later use. Exactly one of args/kwags can be set.
+
+    :param args: at most one unnamed table of the form table_name=table_value
+    :param kwargs: at most one named table of the form table_name=table_value
+    :return: a table description, with all values retained
+    """
+    assert (len(args) + len(kwargs)) == 1
+    if len(kwargs) == 1:
+        table_name = [k for k in kwargs.keys()][0]
+        d = kwargs[table_name]
+        return table(d=d, table_name=table_name)
+    d = args[0]
+    return table(d=d, table_name=None)
+
+
 class ExtendNode(ViewRepresentation):
     def __init__(
         self, *, source, parsed_ops, partition_by=None, order_by=None, reverse=None
@@ -1060,7 +1099,6 @@ class ExtendNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\"
                 + spacer
             )
         ops = [
@@ -1068,7 +1106,7 @@ class ExtendNode(ViewRepresentation):
             for (k, opi) in self.ops.items()
         ]
         flowed = ("," + spacer + " ").join(ops)
-        s = s + ("extend({" + spacer + " " + flowed + "}")
+        s = s + (".extend({" + spacer + " " + flowed + "}")
         if self.windowed_situation:
             if len(self.partition_by) > 0:
                 s = s + "," + spacer + "partition_by=" + self.partition_by.__repr__()
@@ -1083,7 +1121,7 @@ class ExtendNode(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.extend_to_sql(
             self,
             using=using,
@@ -1205,11 +1243,11 @@ class ProjectNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
         s = s + (
-            "project({"
+            ".project({"
             + spacer
             + " "
             + ("," + spacer + " ").join(
@@ -1227,7 +1265,7 @@ class ProjectNode(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.project_to_sql(
             self,
             using=using,
@@ -1289,15 +1327,15 @@ class SelectRowsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("select_rows(" + self.expr.to_python().__repr__() + ")")
+        s = s + (".select_rows(" + self.expr.to_python().__repr__() + ")")
         return s
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.select_rows_to_sql(
             self,
             using=using,
@@ -1363,15 +1401,15 @@ class SelectColumnsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("select_columns(" + self.column_selection.__repr__() + ")")
+        s = s + (".select_columns(" + self.column_selection.__repr__() + ")")
         return s
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.select_columns_to_sql(
             self,
             using=using,
@@ -1437,15 +1475,15 @@ class DropColumnsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("drop_columns(" + self.column_deletions.__repr__() + ")")
+        s = s + (".drop_columns(" + self.column_deletions.__repr__() + ")")
         return s
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.drop_columns_to_sql(
             self,
             using=using,
@@ -1518,10 +1556,10 @@ class OrderRowsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("order_rows(" + self.order_columns.__repr__())
+        s = s + (".order_rows(" + self.order_columns.__repr__())
         if len(self.reverse) > 0:
             s = s + ", reverse=" + self.reverse.__repr__()
         if self.limit is not None:
@@ -1531,7 +1569,7 @@ class OrderRowsNode(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.order_to_sql(
             self,
             using=using,
@@ -1623,15 +1661,15 @@ class RenameColumnsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("rename_columns(" + self.column_remapping.__repr__() + ")")
+        s = s + (".rename_columns(" + self.column_remapping.__repr__() + ")")
         return s
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.rename_to_sql(
             self,
             using=using,
@@ -1727,10 +1765,10 @@ class NaturalJoinNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("natural_join(b=\n" + " " * (indent + 6))
+        s = s + (".natural_join(b=\n" + " " * (indent + 6))
         if print_sources:
             s = s + (
                 self.sources[1].to_python_implementation(
@@ -1748,7 +1786,7 @@ class NaturalJoinNode(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.natural_join_to_sql(
             self,
             using=using,
@@ -1840,10 +1878,10 @@ class ConcatRowsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
-        s = s + ("concat_rows(b=\n" + " " * (indent + 6))
+        s = s + (".concat_rows(b=\n" + " " * (indent + 6))
         if print_sources:
             s = s + (
                 self.sources[1].to_python_implementation(
@@ -1867,7 +1905,7 @@ class ConcatRowsNode(ViewRepresentation):
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         return db_model.concat_rows_to_sql(
             self,
             using=using,
@@ -1916,53 +1954,58 @@ class ConvertRecordsNode(ViewRepresentation):
         if print_sources:
             s = (
                 self.sources[0].to_python_implementation(indent=indent, strict=strict)
-                + " .\\\n"
+                + "\n"
                 + " " * (max(indent, 0) + 3)
             )
         rm_str = self.record_map.__repr__()
         rm_str = re.sub("\n", "\n   ", rm_str)
-        s = s + "convert_records(" + rm_str
+        s = s + ".convert_records(" + rm_str
         s = s + ")"
         return s
 
     def to_near_sql_implementation(
         self, db_model, *, using, temp_id_source, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if temp_id_source is None:
             temp_id_source = [0]
         # TODO: narrow to what we are using
-        sub_query = self.sources[0].to_near_sql_implementation(
+        # TODO: use nearsql instead of strings / lists of strings
+        near_sql = self.sources[0].to_near_sql_implementation(
             db_model=db_model,
             using=None,
             temp_id_source=temp_id_source,
             sql_format_options=sql_format_options,
         )
+        assert isinstance(near_sql, data_algebra.near_sql.NearSQL)
         # claims to use all columns
-        query = sub_query.to_sql(
-            columns=self.columns_used_from_sources()[0],
-            db_model=db_model,
-            force_sql=True,
-        )
         if self.record_map.blocks_in is not None:
-            query = db_model.blocks_to_row_recs_query(
-                query, record_spec=self.record_map.blocks_in
+            view_name = "convert_records_blocks_in_" + str(temp_id_source[0])
+            temp_id_source[0] = temp_id_source[0] + 1
+            pi, si = db_model.blocks_to_row_recs_query_str_list_pair(
+                record_spec=self.record_map.blocks_in
             )
+            near_sql = data_algebra.near_sql.NearSQLRawQStep(
+                prefix=pi,
+                quoted_query_name=db_model.quote_identifier(view_name),
+                sub_sql=data_algebra.near_sql.NearSQLContainer(near_sql=near_sql),
+                suffix=si,
+                annotation="convert records blocks in",
+            )
+            assert isinstance(near_sql, data_algebra.near_sql.NearSQL)
         if self.record_map.blocks_out is not None:
-            query = db_model.row_recs_to_blocks_query(
-                query, record_spec=self.record_map.blocks_out,
+            view_name = "convert_records_blocks_out_" + str(temp_id_source[0])
+            temp_id_source[0] = temp_id_source[0] + 1
+            pi, si = db_model.row_recs_to_blocks_query_str_list_pair(
+                record_spec=self.record_map.blocks_out,
             )
-        view_name = "convert_records_in_" + str(temp_id_source[0])
-        temp_id_source[0] = temp_id_source[0] + 1
-        prev_view_name = "convert_records_out_" + str(temp_id_source[0])
-        temp_id_source[0] = temp_id_source[0] + 1
-        terms = {k: None for k in self.record_map.columns_produced}
-        near_sql = data_algebra.near_sql.NearSQLq(
-            quoted_query_name=db_model.quote_identifier(view_name),
-            prev_quoted_query_name=db_model.quote_identifier(prev_view_name),
-            query=query,
-            terms=terms,
-            annotation="convert records",
-        )
+            near_sql = data_algebra.near_sql.NearSQLRawQStep(
+                prefix=pi,
+                quoted_query_name=db_model.quote_identifier(view_name),
+                sub_sql=data_algebra.near_sql.NearSQLContainer(near_sql=near_sql),
+                suffix=si,
+                annotation="convert records blocks out",
+            )
+            assert isinstance(near_sql, data_algebra.near_sql.NearSQL)
         return near_sql
 
     def eval_implementation(self, *, data_map, data_model, narrow):
@@ -1971,3 +2014,16 @@ class ConvertRecordsNode(ViewRepresentation):
         return data_model.convert_records_step(
             op=self, data_map=data_map, narrow=narrow
         )
+
+
+def ex(d, *, data_model=None, narrow=True, allow_limited_tables=False):
+    """
+    Evaluate operators with respect to Pandas data frames already stored in the operator chain.
+
+    :param d: data algebra pipeline or OpC container to evaluate.
+    :param data_model: adaptor to data dialect (Pandas for now)
+    :param narrow: logical, if True don't copy unexpected columns
+    :param allow_limited_tables: logical, if True allow execution on non-complete tables
+    :return: table result
+    """
+    return d.ex(data_model=data_model, narrow=narrow, allow_limited_tables=allow_limited_tables)
