@@ -180,13 +180,12 @@ class OperatorPlatform:
         raise NotImplementedError("base class called")
 
     def map_records(
-        self, blocks_in=None, blocks_out=None, strict=False
-    ):
+        self, blocks_in=None, blocks_out=None):
         if (blocks_in is None) and (blocks_out is None):
             return self  # NO-OP, return source ops
         return self.convert_records(
             data_algebra.cdata.RecordMap(
-                blocks_in=blocks_in, blocks_out=blocks_out, strict=strict
+                blocks_in=blocks_in, blocks_out=blocks_out
             ),
         )
 
