@@ -1,7 +1,10 @@
+
 import math
 import re
 from collections import OrderedDict
+from typing import List, Optional
 from types import SimpleNamespace
+from typing import Tuple
 
 import pandas.io.sql
 
@@ -11,14 +14,28 @@ import data_algebra.near_sql
 import data_algebra.expr_rep
 import data_algebra.util
 import data_algebra.data_ops_types
-import data_algebra.eval_model
 import data_algebra.data_ops
 
 
+# The db_model can be a bit tricky as SQL is represented a few ways depending
+# on how close to a final result we are.
+# The end representation is a single string.
+# The nearly there representation is a list of strings, which makes indenting much easier.
+# The primary computational representation is a NearSQL structure, as it is a dag of objects.
+# Note: near sql has a bound/unbound variation treating the top layer differently than
+# subordinate nodes.
+
 class SQLFormatOptions(SimpleNamespace):
+    """
+    Simple class for holding SQL formatting options
+    """
     def __init__(
-        self, use_with=True, annotate=True, sql_indent=" ", initial_commas=False,
+        self, use_with: bool = True, annotate: bool = True, sql_indent: str = " ", initial_commas: bool = False,
     ):
+        assert isinstance(use_with, bool)
+        assert isinstance(annotate, bool)
+        assert isinstance(sql_indent, str)
+        assert isinstance(initial_commas, bool)
         SimpleNamespace.__init__(
             self,
             use_with=use_with,
@@ -27,21 +44,32 @@ class SQLFormatOptions(SimpleNamespace):
             initial_commas=initial_commas,
         )
 
+    def __str__(self):
+        return self.__repr__()
 
-def _str_join_expecting_list(joiner, str_list):
+    def _repr_pretty_(self, p, cycle):
+        """
+        IPython pretty print, used at implicit print time
+        https://ipython.readthedocs.io/en/stable/config/integrating.html
+        """
+        p.text(str(self))
+
+
+def _list_join_expecting_list(joiner: str, str_list: List[str]):
     assert isinstance(joiner, str)
     assert isinstance(str_list, list)
     assert all([isinstance(vi, str) for vi in str_list])
-    return joiner.join(str_list)
+    n = len(str_list)
+    return [' ' + str_list[i] + (joiner if i < (n - 1) else '') for i in range(n)]
 
 
-def _clean_annotation(annotation):
+def _clean_annotation(annotation: Optional[str]):
     assert isinstance(annotation, (str, type(None)))
     if annotation is None:
         return annotation
     annotation = annotation.strip()
-    annotation = re.sub(r"\s+", " ", annotation)
-    return annotation
+    annotation = re.sub(r"(\s|\r|\n)+", " ", annotation)
+    return annotation.strip()
 
 
 # map from op-name to special SQL formatting code
@@ -475,7 +503,6 @@ class DBModel:
     on_joiner: str
     drop_text: str
     string_type: str
-    join_name_map: dict
     supports_with: bool
     allow_extend_merges: bool
     default_SQL_format_options: SQLFormatOptions
@@ -495,7 +522,6 @@ class DBModel:
         on_joiner="AND",
         drop_text="DROP TABLE",
         string_type="VARCHAR",
-        join_name_map=None,
         supports_with=True,
         allow_extend_merges=True,
         default_SQL_format_options=None,
@@ -522,9 +548,6 @@ class DBModel:
         self.on_joiner = on_joiner
         self.drop_text = drop_text
         self.string_type = string_type
-        if join_name_map is None:
-            join_name_map = {}
-        self.join_name_map = join_name_map.copy()
         if default_SQL_format_options is None:
             default_SQL_format_options = SQLFormatOptions()
         assert isinstance(default_SQL_format_options, SQLFormatOptions)
@@ -553,6 +576,7 @@ class DBModel:
             q = q.to_sql(db_model=self)
         else:
             q = str(q)
+        assert isinstance(q, str)
         pandas.io.sql.execute(q, conn)
 
     def read_query(self, conn, q):
@@ -566,6 +590,7 @@ class DBModel:
             q = q.to_sql(db_model=self)
         else:
             q = str(q)
+        assert isinstance(q, str)
         r = pandas.io.sql.read_sql(q, conn)
         r = self.local_data_model.pd.DataFrame(r)
         r = r.reset_index(drop=True)
@@ -688,7 +713,7 @@ class DBModel:
             return "(" + ", ".join([self.value_to_sql(vi) for vi in v]) + ")"
         return str(v)
 
-    def table_values_to_sql(self, v):
+    def table_values_to_sql_str_list(self, v) -> str:
         assert v is not None
         m = v.shape[0]
         assert m > 0
@@ -715,7 +740,7 @@ class DBModel:
             + ["    " + ("" if (i < 1) else "UNION ALL ") + q_row(i) for i in range(m)]
             + [f') {qi("table_values")}']
         )
-        return "\n".join(sql) + "\n"
+        return sql
 
     def expr_to_sql(self, expression, *, want_inline_parens=False):
         if isinstance(expression, str):
@@ -778,7 +803,7 @@ class DBModel:
 
     def table_def_to_sql(
         self, table_def, *, using=None, temp_id_source=None, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if table_def.node_name != "TableDescription":
             raise TypeError(
                 "Expected table_def to be a data_algebra.data_ops.TableDescription)"
@@ -816,7 +841,7 @@ class DBModel:
 
     def extend_to_sql(
         self, extend_node, *, using=None, temp_id_source=None, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if extend_node.node_name != "ExtendNode":
             raise TypeError(
                 "Expected extend_node to be a data_algebra.data_ops.ExtendNode)"
@@ -873,16 +898,16 @@ class DBModel:
         origcols = [k for k in using if k not in subops.keys()]
         for k in origcols:
             terms[k] = None
-            declared_term_dependencies[k] = set([k])
+            declared_term_dependencies[k] = {k}
         for (ci, oi) in subops.items():
             terms[ci] = self.expr_to_sql(oi) + window_term
             cols_used_in_term = set()
             oi.get_column_names(cols_used_in_term)
             cols_used_in_term.update(window_vars)
             declared_term_dependencies[ci] = cols_used_in_term
-        annotation = extend_node.to_python_implementation(
+        annotation = str(extend_node.to_python_implementation(
             print_sources=False, indent=-1
-        )
+        ))
         # TODO: see if we can merge with subsql instead of building a new one
         if (
             self.allow_extend_merges
@@ -894,11 +919,11 @@ class DBModel:
             # check detailed merge conditions
             def non_trivial_terms(*, dep_dict, term_dict):
                 return [
-                    k
-                    for k, v in dep_dict.items()
-                    if (len(v - set([k])) > 0)
-                    or (k not in v)
-                    or ((term_dict[k] is not None) and (term_dict[k] != k))
+                    ki
+                    for ki, vi in dep_dict.items()
+                    if (len(vi - {ki}) > 0)
+                    or (ki not in vi)
+                    or ((term_dict[ki] is not None) and (term_dict[ki] != ki))
                 ]
 
             our_non_trivial_terms = non_trivial_terms(
@@ -979,9 +1004,9 @@ class DBModel:
             quoted_query_name=self.quote_identifier(view_name),
             sub_sql=subsql.to_bound_near_sql(columns=subusing),
             suffix=suffix,
-            annotation=project_node.to_python_implementation(
+            annotation=str(project_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
@@ -992,7 +1017,7 @@ class DBModel:
         using=None,
         temp_id_source=None,
         sql_format_options=None,
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if select_rows_node.node_name != "SelectRowsNode":
             raise TypeError(
                 "Expected select_rows_node to be a data_algebra.data_ops.SelectRowsNode)"
@@ -1019,9 +1044,9 @@ class DBModel:
             quoted_query_name=self.quote_identifier(view_name),
             sub_sql=subsql.to_bound_near_sql(columns=subusing),
             suffix=suffix,
-            annotation=select_rows_node.to_python_implementation(
+            annotation=str(select_rows_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
@@ -1032,7 +1057,7 @@ class DBModel:
         using=None,
         temp_id_source=None,
         sql_format_options=None,
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if select_columns_node.node_name != "SelectColumnsNode":
             raise TypeError(
                 "Expected select_columns_to_sql to be a data_algebra.data_ops.SelectColumnsNode)"
@@ -1049,11 +1074,14 @@ class DBModel:
             db_model=self, using=set(subusing), temp_id_source=temp_id_source
         )
         # order/limit columns
-        subsql.terms = {
-            k: subsql.terms[k]
-            for k in select_columns_node.column_selection
-            if k in subusing
-        }
+        if subsql.terms is not None:
+            subsql.terms = {
+                k: subsql.terms[k]
+                for k in select_columns_node.column_selection
+                if k in subusing
+            }
+        else:
+            subsql.terms = []
         return subsql
 
     def drop_columns_to_sql(
@@ -1063,7 +1091,7 @@ class DBModel:
         using=None,
         temp_id_source=None,
         sql_format_options=None,
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if drop_columns_node.node_name != "DropColumnsNode":
             raise TypeError(
                 "Expected drop_columns_node to be a data_algebra.data_ops.DropColumnsNode)"
@@ -1086,7 +1114,7 @@ class DBModel:
 
     def order_to_sql(
         self, order_node, *, using=None, temp_id_source=None, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if order_node.node_name != "OrderRowsNode":
             raise TypeError(
                 "Expected order_node to be a data_algebra.data_ops.OrderRowsNode)"
@@ -1128,15 +1156,15 @@ class DBModel:
             quoted_query_name=self.quote_identifier(view_name),
             sub_sql=subsql.to_bound_near_sql(columns=subusing),
             suffix=suffix,
-            annotation=order_node.to_python_implementation(
+            annotation=str(order_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
     def rename_to_sql(
         self, rename_node, *, using=None, temp_id_source=None, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if rename_node.node_name != "RenameColumnsNode":
             raise TypeError(
                 "Expected rename_node to be a data_algebra.data_ops.RenameColumnsNode)"
@@ -1163,13 +1191,13 @@ class DBModel:
             terms=terms,
             quoted_query_name=self.quote_identifier(view_name),
             sub_sql=subsql.to_bound_near_sql(columns=subusing),
-            annotation=rename_node.to_python_implementation(
+            annotation=str(rename_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
-    def _coalesce_terms(self, *, sub_view_name_left, sub_view_name_right, cols):
+    def _coalesce_terms(self, *, sub_view_name_first, sub_view_name_second, cols):
         coalesce_formatter = self.sql_formatters["coalesce"]
 
         class PseudoExpression:
@@ -1181,8 +1209,8 @@ class DBModel:
                 self,
                 PseudoExpression(
                     [
-                        sub_view_name_left + "." + self.quote_identifier(ci),
-                        sub_view_name_right + "." + self.quote_identifier(ci),
+                        sub_view_name_first + "." + self.quote_identifier(ci),
+                        sub_view_name_second + "." + self.quote_identifier(ci),
                     ]
                 ),
             )
@@ -1190,53 +1218,66 @@ class DBModel:
         }
         return terms
 
-    def natural_join_to_sql(
-        self, join_node, *, using=None, temp_id_source=None, sql_format_options=None
+    def _natural_join_sub_queries(
+            self,  *, join_node, using, temp_id_source
     ):
         if join_node.node_name != "NaturalJoinNode":
             raise TypeError(
                 "Expected join_node to be a data_algebra.data_ops.NaturalJoinNode)"
             )
-        if temp_id_source is None:
-            temp_id_source = [0]
         if using is None:
             using = join_node.column_set
         by_set = set(join_node.by)
         if len(using) < 1:
-            raise ValueError("must select at least one column")
+            raise ValueError("join must use or select at least one column")
         missing = using - join_node.column_set
         if len(missing) > 0:
             raise KeyError("referred to unknown columns: " + str(missing))
-        subusing = join_node.columns_used_from_sources(using=using.union(by_set))
-        using_left = subusing[0]
-        using_right = subusing[1]
+        using_left, using_right = join_node.columns_used_from_sources(using=using.union(by_set))
         sql_left = join_node.sources[0].to_near_sql_implementation(
             db_model=self, using=using_left, temp_id_source=temp_id_source
         )
-        sub_view_name_left = sql_left.quoted_query_name
         sql_right = join_node.sources[1].to_near_sql_implementation(
             db_model=self, using=using_right, temp_id_source=temp_id_source
         )
-        sub_view_name_right = sql_right.quoted_query_name
+        return using_left, sql_left, using_right, sql_right
+
+    def natural_join_to_sql(
+        self, join_node, *, using=None, temp_id_source=None, sql_format_options=None, left_is_first=True
+    ) -> data_algebra.near_sql.NearSQL:
+        if temp_id_source is None:
+            temp_id_source = [0]
+        if using is None:
+            using = join_node.column_set
+        using_left, sql_left, using_right, sql_right = self._natural_join_sub_queries(
+            join_node=join_node, using=using, temp_id_source=temp_id_source
+        )
         view_name = "natural_join_" + str(temp_id_source[0])
         temp_id_source[0] = temp_id_source[0] + 1
         common = using_left.intersection(using_right)
-        terms = self._coalesce_terms(
-            sub_view_name_left=sub_view_name_left,
-            sub_view_name_right=sub_view_name_right,
-            cols=[ci for ci in common if ci in using],
-        )
+        if left_is_first:
+            terms = self._coalesce_terms(
+                sub_view_name_first=sql_left.quoted_query_name,
+                sub_view_name_second=sql_right.quoted_query_name,
+                cols=[ci for ci in common if ci in using],
+            )
+        else:
+            terms = self._coalesce_terms(
+                sub_view_name_first=sql_right.quoted_query_name,
+                sub_view_name_second=sql_left.quoted_query_name,
+                cols=[ci for ci in common if ci in using],
+            )
         terms.update({ci: None for ci in using_left - common})
         terms.update({ci: None for ci in using_right - common})
         on_terms = []
         if len(join_node.by) > 0:
             on_terms = ["ON " + self.on_start] + self._indent_and_sep_terms(
                 [
-                    sub_view_name_left
+                    sql_left.quoted_query_name
                     + "."
                     + self.quote_identifier(c)
                     + " = "
-                    + sub_view_name_right
+                    + sql_right.quoted_query_name
                     + "."
                     + self.quote_identifier(c)
                     for c in join_node.by
@@ -1246,29 +1287,22 @@ class DBModel:
             )
             if (self.on_end is not None) and (len(self.on_end) > 0):
                 on_terms = on_terms + [self.on_end]
-        jointype = join_node.jointype
-        try:
-            jointype = self.join_name_map[
-                jointype
-            ]  # TODO: maybe move this mapping earlier
-        except KeyError:
-            pass
         near_sql = data_algebra.near_sql.NearSQLBinaryStep(
             terms=terms,
             quoted_query_name=self.quote_identifier(view_name),
             sub_sql1=sql_left.to_bound_near_sql(columns=using_left, force_sql=False),
-            joiner=jointype + " JOIN",
+            joiner=join_node.jointype + " JOIN",
             sub_sql2=sql_right.to_bound_near_sql(columns=using_right, force_sql=False),
             suffix=on_terms,
-            annotation=join_node.to_python_implementation(
+            annotation=str(join_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
     def concat_rows_to_sql(
         self, concat_node, *, using=None, temp_id_source=None, sql_format_options=None
-    ):
+    ) -> data_algebra.near_sql.NearSQL:
         if concat_node.node_name != "ConcatRowsNode":
             raise TypeError(
                 "Expected join_node to be a data_algebra.data_ops.ConcatRowsNode)"
@@ -1316,15 +1350,15 @@ class DBModel:
             sub_sql2=sql_right.to_bound_near_sql(
                 columns=using_left, force_sql=True, constants=constants_right,
             ),
-            annotation=concat_node.to_python_implementation(
+            annotation=str(concat_node.to_python_implementation(
                 print_sources=False, indent=-1
-            ),
+            )),
         )
         return near_sql
 
     def to_sql(
         self, ops, *, sql_format_options=None,
-    ):
+    ) -> str:
         assert isinstance(self, DBModel)
         assert isinstance(ops, data_algebra.data_ops.ViewRepresentation)
         if sql_format_options is None:
@@ -1339,13 +1373,13 @@ class DBModel:
         sql_str_list = None
         if sql_format_options.use_with and self.supports_with:
             sequence = near_sql.to_with_form()
-            len_sequence = len(sequence)
+            len_sequence = len(sequence.previous_steps)
             # can fall back to the non-with path
-            if len(sequence) >= 2:
+            if len(sequence.previous_steps) >= 1:
                 sql_sequence = []
-                for i in range(len_sequence - 1):
-                    nmi = sequence[i][0]  # already quoted
-                    sqli = sequence[i][1].to_sql(
+                for i in range(len_sequence):
+                    nmi = sequence.previous_steps[i][0]  # already quoted
+                    sqli = sequence.previous_steps[i][1].to_sql_str_list(
                         db_model=self, sql_format_options=sql_format_options
                     )
                     sql_sequence = (
@@ -1358,21 +1392,21 @@ class DBModel:
                             for s in sqli
                         ]
                     )
-                    if i < (len_sequence - 2):
+                    if i < (len_sequence - 1):
                         sql_sequence = sql_sequence + [
-                            sql_format_options.sql_indent + "),"
+                            sql_format_options.sql_indent + ") ,"
                         ]
                     else:
                         sql_sequence = sql_sequence + [
                             sql_format_options.sql_indent + ")"
                         ]
-                sql_last = sequence[len_sequence - 1].to_sql(
+                sql_last = sequence.last_step.to_sql_str_list(
                     db_model=self, force_sql=True, sql_format_options=sql_format_options
                 )
                 sql_str_list = ["WITH"] + sql_sequence + sql_last
         if sql_str_list is None:
             # non-with path
-            sql_str_list = near_sql.to_sql(
+            sql_str_list = near_sql.to_sql_str_list(
                 db_model=self, force_sql=True, sql_format_options=sql_format_options
             )
         if sql_format_options.annotate:
@@ -1383,19 +1417,12 @@ class DBModel:
                 f"--       string quote: {self.string_quote}",
                 f"--   identifier quote: {self.identifier_quote}",
             ] + sql_str_list
+        sql_str_list = [v.rstrip() for v in sql_str_list]
+        sql_str_list = [v for v in sql_str_list if len(v) > 0]
         return "\n".join(sql_str_list) + "\n"
 
-    def row_recs_to_blocks_query(
-        self, source_sql, record_spec, *, using=None, temp_id_source=None
-    ):
-        if temp_id_source is None:
-            temp_id_source = [0]
-        if isinstance(source_sql, str):
-            source_sql = [source_sql]
-        # if not isinstance(record_spec, data_algebra.cdata.RecordSpecification):
-        #     raise TypeError(
-        #         "record_spec should be a data_algebra.cdata.RecordSpecification"
-        #     )
+    def row_recs_to_blocks_query_str_list_pair(
+        self, record_spec) -> Tuple[List[str], List[str]]:
         control_value_cols = [
             c
             for c in record_spec.control_table.columns
@@ -1424,7 +1451,7 @@ class DBModel:
             if result_col in seen:
                 continue
             seen.add(result_col)
-            cstmt = " CASE\n"
+            cstmt = " CASE "
             col = record_spec.control_table[result_col]
             isnull = col.isnull()
             for i in range(len(col)):
@@ -1439,57 +1466,57 @@ class DBModel:
                         + self.quote_string(str(source_col))
                         + " THEN a."
                         + self.quote_identifier(source_col)
-                        + "\n"
+                        + " "
                     )
                     cstmt = cstmt + col_sql
-            cstmt = cstmt + "  ELSE NULL END AS " + self.quote_identifier(result_col)
+            cstmt = cstmt + " ELSE NULL END AS " + self.quote_identifier(result_col)
             col_stmts.append(cstmt)
-        ctab_sql = [self.table_values_to_sql(record_spec.control_table)]
-        sql = (
-            "SELECT\n"
-            + _str_join_expecting_list(",\n", col_stmts)
-            + "\n"
-            + "FROM (\n  "
-            + _str_join_expecting_list("\n", source_sql)
-            + " ) a\n"
-            + "CROSS JOIN (\n  "
-            + _str_join_expecting_list("\n", ctab_sql)
-            + " ) b\n"
-            + " ORDER BY "
-            + _str_join_expecting_list(", ", control_cols)
+        ctab_sql = self.table_values_to_sql_str_list(record_spec.control_table)
+        sql_prefix = (
+            _list_join_expecting_list(",", col_stmts)
+            + ["FROM ( SELECT * FROM "]
         )
-        return sql
+        sql_suffix = (
+            [" ) a"]
+            + ["CROSS JOIN ("]
+            + _list_join_expecting_list("", ctab_sql)
+            + [" ) b"]
+            + [" ORDER BY"]  # order by not required, but nice to have
+            + _list_join_expecting_list(", ", control_cols)
+        )
+        return sql_prefix, sql_suffix
 
     # noinspection PyUnusedLocal
-    def blocks_to_row_recs_query(
+    def blocks_to_row_recs_query_str_list_pair(
         self,
-        source_sql,
-        record_spec,
-        *,
-        using=None,
-        temp_id_source=None,
-        sql_format_options=None,
-    ):
-        # if not isinstance(record_spec, data_algebra.cdata.RecordSpecification):
-        #     raise TypeError(
-        #         "record_spec should be a data_algebra.cdata.RecordSpecification"
-        #     )
-        if sql_format_options is None:
-            sql_format_options = self.default_SQL_format_options
-        assert isinstance(sql_format_options, SQLFormatOptions)
-        if isinstance(source_sql, str):
-            source_sql = [source_sql]
+        record_spec) -> Tuple[List[str], List[str]]:
+        assert record_spec.control_table.shape[0] >= 1
+        col_stmts = []
+        for c in record_spec.record_keys:
+            col_stmts.append(
+                " " + self.quote_identifier(c) + " AS " + self.quote_identifier(c)
+            )
+        if record_spec.control_table.shape[0] == 1:
+            # special case with one row control table (rowrec)
+            for cc in record_spec.control_table.columns:
+                cc0 = record_spec.control_table[cc][0]
+                col_stmts.append(
+                    " " + self.quote_identifier(cc) + " AS " + self.quote_identifier(cc0)
+                )
+            sql_prefix = (
+                    _list_join_expecting_list(',', col_stmts)
+                    + ["FROM ( SELECT * FROM "]
+            )
+            sql_suffix = (
+                [" ) a"]
+            )
+            return sql_prefix, sql_suffix
         control_value_cols = [
             c
             for c in record_spec.control_table.columns
             if c not in record_spec.control_table_keys
         ]
         control_cols = [self.quote_identifier(c) for c in record_spec.record_keys]
-        col_stmts = []
-        for c in record_spec.record_keys:
-            col_stmts.append(
-                " " + self.quote_identifier(c) + " AS " + self.quote_identifier(c)
-            )
         seen = set()
         for i in range(record_spec.control_table.shape[0]):
             for vc in control_value_cols:
@@ -1517,24 +1544,21 @@ class DBModel:
                         + self.quote_identifier(col[i])
                     )
                     col_stmts.append(cstmt)
-        sql = (
-            "SELECT\n"
-            + _str_join_expecting_list(",\n", col_stmts)
-            + "\n"
-            + "FROM (\n  "
-            + _str_join_expecting_list("\n", source_sql)
-            + "\n"
-            + " ) a\n"
-            + " GROUP BY "
-            + _str_join_expecting_list(", ", control_cols)
-            + "\n"
-            + " ORDER BY "
-            + _str_join_expecting_list(", ", control_cols)
+        sql_prefix = (
+                _list_join_expecting_list(',', col_stmts)
+                + ["FROM ( SELECT * FROM "]
         )
-        return sql
+        sql_suffix = (
+                [" ) a"]
+                + ["GROUP BY"]
+                + _list_join_expecting_list(',', control_cols)
+                + ["ORDER BY "]  # order by not required, but nice to have
+                + _list_join_expecting_list(',', control_cols)
+        )
+        return sql_prefix, sql_suffix
 
     # encode and name a term for use in a SQL expression
-    def enc_term_(self, k, *, terms):
+    def enc_term_(self, k, *, terms) -> str:
         v = None
         try:
             v = terms[k]
@@ -1546,12 +1570,12 @@ class DBModel:
 
     def convert_nearsql_container_subsql_(
         self, nearsql_container, *, sql_format_options=None
-    ):
+    ) -> List[str]:
         assert isinstance(nearsql_container, data_algebra.near_sql.NearSQLContainer)
         if sql_format_options is None:
             sql_format_options = self.default_SQL_format_options
         assert isinstance(sql_format_options, SQLFormatOptions)
-        sub_sql = nearsql_container.to_sql(self, sql_format_options=sql_format_options)
+        sub_sql = nearsql_container.to_sql_str_list(self, sql_format_options=sql_format_options)
         assert isinstance(sub_sql, list)
         if isinstance(nearsql_container.near_sql, data_algebra.near_sql.NearSQLTable):
             sql = sub_sql
@@ -1571,7 +1595,7 @@ class DBModel:
             )
         return sql
 
-    def nearsqlcte_to_sql_(
+    def nearsqlcte_to_sql_str_list_(
         self,
         near_sql,
         *,
@@ -1579,7 +1603,7 @@ class DBModel:
         force_sql=False,
         constants=None,
         sql_format_options=None,
-    ):
+    ) -> List[str]:
         assert isinstance(near_sql, data_algebra.near_sql.NearSQLCommonTableExpression)
         if sql_format_options is None:
             sql_format_options = self.default_SQL_format_options
@@ -1593,7 +1617,7 @@ class DBModel:
             ]
         return [near_sql.quoted_query_name]
 
-    def nearsqltable_to_sql_(
+    def nearsqltable_to_sql_str_list_(
         self,
         near_sql,
         *,
@@ -1601,15 +1625,18 @@ class DBModel:
         force_sql=False,
         constants=None,
         sql_format_options=None,
-    ):
+    ) -> List[str]:
         assert isinstance(near_sql, data_algebra.near_sql.NearSQLTable)
         if sql_format_options is None:
             sql_format_options = self.default_SQL_format_options
         assert isinstance(sql_format_options, SQLFormatOptions)
         if columns is None:
-            columns = [k for k in near_sql.terms.keys()]
+            if near_sql.terms is not None:
+                columns = [k for k in near_sql.terms.keys()]
+            else:
+                columns = []
         if len(columns) <= 0:
-            force_sql = False
+            columns = []
         have_constants = (constants is not None) and (len(constants) > 0)
         if force_sql or have_constants:
             terms_strs = [self.quote_identifier(k) for k in columns]
@@ -1620,7 +1647,7 @@ class DBModel:
                 ]
             if len(terms_strs) < 1:
                 terms_strs = [
-                    f'1 AS {self.quote_identifier("data_algebra_placeholder_col_name")}'
+                    '*'
                 ]
             return (
                 ["SELECT"]
@@ -1632,7 +1659,7 @@ class DBModel:
             )
         return [near_sql.quoted_table_name]
 
-    def nearsqlunary_to_sql_(
+    def nearsqlunary_to_sql_str_list_(
         self,
         near_sql,
         *,
@@ -1640,7 +1667,7 @@ class DBModel:
         force_sql=False,
         constants=None,
         sql_format_options=None,
-    ):
+    ) -> List[str]:
         assert isinstance(near_sql, data_algebra.near_sql.NearSQLUnaryStep)
         if sql_format_options is None:
             sql_format_options = self.default_SQL_format_options
@@ -1658,7 +1685,7 @@ class DBModel:
                 ]
             if len(terms_strs) < 1:
                 terms_strs = [
-                    f'1 AS {self.quote_identifier("data_algebra_placeholder_col_name")}'
+                    '*'
                 ]
         sql_start = "SELECT"
         if (
@@ -1684,7 +1711,34 @@ class DBModel:
             sql = sql + near_sql.suffix
         return sql
 
-    def nearsqlbinary_to_sql_(
+    def nearsqlrawq_to_sql_str_list_(
+        self,
+        near_sql,
+        *,
+        columns=None,
+        force_sql=False,
+        constants=None,
+        sql_format_options=None,
+    ) -> List[str]:
+        assert isinstance(near_sql, data_algebra.near_sql.NearSQLRawQStep)
+        if sql_format_options is None:
+            sql_format_options = self.default_SQL_format_options
+        assert isinstance(sql_format_options, SQLFormatOptions)
+        sql = (
+            ["SELECT"]
+            + [' ' + v for v in near_sql.prefix]
+            + [
+                sql_format_options.sql_indent + si
+                for si in near_sql.sub_sql.convert_subsql(
+                    db_model=self, sql_format_options=sql_format_options
+                )
+            ]
+        )
+        if (near_sql.suffix is not None) and (len(near_sql.suffix) > 0):
+            sql = sql + [' ' + v for v in near_sql.suffix]
+        return sql
+
+    def nearsqlbinary_to_sql_str_list_(
         self,
         near_sql,
         *,
@@ -1693,7 +1747,7 @@ class DBModel:
         constants=None,
         sql_format_options=None,
         quoted_query_name=None,
-    ):
+    ) -> List[str]:
         assert isinstance(near_sql, data_algebra.near_sql.NearSQLBinaryStep)
         if sql_format_options is None:
             sql_format_options = self.default_SQL_format_options
@@ -1706,7 +1760,7 @@ class DBModel:
         terms_strs = [self.enc_term_(k, terms=terms) for k in columns]
         if len(terms_strs) < 1:
             terms_strs = [
-                f'1 AS {self.quote_identifier("data_algebra_placeholder_col_name")}'
+                '*'
             ]
         is_union = "union" in near_sql.joiner.lower()
         sql_start = "SELECT"
@@ -1717,10 +1771,10 @@ class DBModel:
         ):
             sql_start = "SELECT  -- " + _clean_annotation(near_sql.annotation)
         if is_union:
-            substr_1 = near_sql.sub_sql1.to_sql(
+            substr_1 = near_sql.sub_sql1.to_sql_str_list(
                 db_model=self, sql_format_options=sql_format_options
             )
-            substr_2 = near_sql.sub_sql2.to_sql(
+            substr_2 = near_sql.sub_sql2.to_sql_str_list(
                 db_model=self, sql_format_options=sql_format_options
             )
         else:
@@ -1753,48 +1807,6 @@ class DBModel:
             sql = sql + [")"]
         return sql
 
-    def nearsqlq_to_sql_(
-        self, near_sql, *, columns=None, constants=None, sql_format_options=None
-    ):
-        assert isinstance(near_sql, data_algebra.near_sql.NearSQLq)
-        if sql_format_options is None:
-            sql_format_options = self.default_SQL_format_options
-        assert isinstance(sql_format_options, SQLFormatOptions)
-        if columns is None:
-            columns = [k for k in near_sql.terms.keys()]
-        terms = near_sql.terms
-        if (constants is not None) and (len(constants) > 0):
-            terms.update(constants)
-
-        def enc_term(k):
-            v = terms[k]
-            if v is None:
-                return self.quote_identifier(k)
-            return v + " AS " + self.quote_identifier(k)
-
-        terms_strs = [enc_term(k) for k in columns]
-        if len(terms_strs) < 1:
-            terms_strs = [
-                f'1 AS {self.quote_identifier("data_algebra_placeholder_col_name")}'
-            ]
-        sql_start = "SELECT"
-        if (
-            sql_format_options.annotate
-            and (near_sql.annotation is not None)
-            and (len(near_sql.annotation) > 0)
-        ):
-            sql_start = "SELECT  -- " + _clean_annotation(near_sql.annotation) + "\n "
-        return (
-            [sql_start]
-            + self._indent_and_sep_terms(
-                terms_strs, sql_format_options=sql_format_options
-            )
-            + ["FROM"]
-            + ["("]
-            + [near_sql.query]  # TODO: see if we can vectorized
-            + [") " + near_sql.prev_quoted_query_name]
-        )
-
     def __str__(self):
         return str(type(self).__name__)
 
@@ -1802,10 +1814,9 @@ class DBModel:
         return self.__str__()
 
 
-class DBHandle(data_algebra.eval_model.EvalModel):
+class DBHandle:
     def __init__(self, *, db_model, conn):
         assert isinstance(db_model, DBModel)
-        data_algebra.eval_model.EvalModel.__init__(self)
         self.db_model = db_model
         self.conn = conn
 
@@ -1863,51 +1874,15 @@ class DBHandle(data_algebra.eval_model.EvalModel):
 
     def to_sql(
         self, ops, *, sql_format_options=None,
-    ):
+    ) -> str:
         return self.db_model.to_sql(ops=ops, sql_format_options=sql_format_options,)
 
     def query_to_csv(self, q, *, res_name):
         d = self.read_query(q)
         d.to_csv(res_name, index=False)
 
-    def table_values_to_sql(self, v):
-        return self.db_model.table_values_to_sql(v)
-
-    def managed_eval(self, ops, *, data_map=None, result_name=None, narrow=True):
-        query = ops.to_sql(self.db_model)
-        if result_name is None:
-            result_name = self.mk_tmp_name(data_map)
-        tables_needed = [k for k in ops.get_tables().keys()]
-        if data_map is not None:
-            missing_tables = set(tables_needed) - set(data_map.keys())
-            if len(missing_tables) > 0:
-                raise ValueError("missing required tables: " + str(missing_tables))
-            for k in tables_needed:
-                assert isinstance(data_map[k], data_algebra.data_ops.TableDescription)
-                if data_map[k].table_name != k:
-                    raise ValueError(
-                        "data_map["
-                        + k
-                        + "].table_name == "
-                        + data_map[k].table_name
-                        + ", not "
-                        + k
-                    )
-        if result_name in tables_needed:
-            raise ValueError("Can not write over an input table")
-        q_table_name = self.db_model.quote_table_name(result_name)
-        drop_query = self.db_model.drop_text + " " + q_table_name
-        create_query = "CREATE TABLE " + q_table_name + " AS " + query
-        # noinspection PyBroadException
-        try:
-            self.db_model.execute(self.conn, drop_query)
-        except Exception:
-            pass
-        self.db_model.execute(self.conn, create_query)
-        res = self.describe_table(result_name)
-        if data_map is not None:
-            data_map[result_name] = res
-        return result_name
+    def table_values_to_sql_str_list(self, v) -> str:
+        return self.db_model.table_values_to_sql_str_list(v)
 
     def __str__(self):
         return (
