@@ -14,6 +14,24 @@ import data_algebra.util
 #  http://tomerfiliba.com/blog/Infix-Operators/
 
 
+class PythonText:
+    """
+    Class for holding text representation of Python, with possible additional annotations.
+    str() method returns only the text for interoperability.
+    """
+    def __init__(self, s: str, *, is_in_parens: bool = False):
+        assert isinstance(s, str)
+        assert isinstance(is_in_parens, bool)
+        self.s = s
+        self.is_in_parens = is_in_parens
+
+    def __str__(self):
+        return self.s
+
+    def __repr__(self):
+        return self.s.__repr__()
+
+
 # list of window/aggregation functions that must be windowed/aggregated
 # (note some other functions work in more than one mode)
 # noinspection SpellCheckingInspection
@@ -130,7 +148,12 @@ class PreTerm(ABC):
 
     # emitters
 
-    def to_python(self, *, want_inline_parens=False):
+    def to_python(self, *, want_inline_parens : bool = False) -> PythonText:
+        """
+        Convert parsed expression into a string
+
+        :param want_inline_parens: bool if True put parens around complex expressions that don't already have a grouper.
+        """
         raise NotImplementedError(
             "base class method called"
         )  # https://docs.python.org/3/library/exceptions.html
@@ -144,10 +167,51 @@ class PreTerm(ABC):
     # printing
 
     def __repr__(self):
-        return self.to_python(want_inline_parens=False)
+        return str(self.to_python(want_inline_parens=False))
 
     def __str__(self):
-        return self.to_python(want_inline_parens=False)
+        return str(self.to_python(want_inline_parens=False))
+
+
+def _is_none_value(x):
+    if x is None:
+        return True
+    if isinstance(x, Value):
+        return x.value is None
+    return False
+
+
+def _check_expr_incompatible_types(a, b):
+    """
+    return None if no type problem, else return pair of types
+    """
+
+    def obvious_declared_type(v):
+        if v is None:
+            return None  # None may be a placeholder for "don't know at this point"
+        if isinstance(v, Value):
+            return type(v.value)
+        if isinstance(v, ColumnReference):
+            if (v.view is None) or (v.view.column_types is None) or (len(v.view.column_types) <= 0):
+                return None
+            try:
+                return v.view.column_types[v.column_name]
+            except KeyError:
+                return None
+        if isinstance(v, PreTerm):
+            return None
+        return None  # dunno
+
+    type_a = obvious_declared_type(a)
+    if type_a is None:
+        return None
+    type_b = obvious_declared_type(b)
+    if type_b is None:
+        return None
+    if data_algebra.util.compatible_types([type_a, type_b]):
+        return None
+    looks_compatible = data_algebra.util.compatible_types([type_a, type_b])
+    return (type_a, type_b)
 
 
 # noinspection SpellCheckingInspection
@@ -161,11 +225,20 @@ class Term(PreTerm, ABC):
 
     # builders
 
-    def __op_expr__(self, op, other, *, inline=True, method=False):
+    def __op_expr__(self, op, other, *, inline=True, method=False, check_types=True):
         """binary expression"""
         assert isinstance(op, str)
         if not isinstance(other, Term):
             other = enc_value(other)
+        assert not _is_none_value(self)
+        assert not _is_none_value(other)
+        if check_types:
+            obvious_type_problem = _check_expr_incompatible_types(self, other)
+            if obvious_type_problem is not None:
+                raise TypeError(
+                    f"trying to combine incompatible values:"
+                    + f" {self}:{obvious_type_problem[0]}"
+                    + f" and {other}:{obvious_type_problem[0]} with {op}")
         return Expression(op, (self, other), inline=inline, method=method)
 
     def __rop_expr__(self, op, other):
@@ -173,11 +246,14 @@ class Term(PreTerm, ABC):
         assert isinstance(op, str)
         if not isinstance(other, Term):
             other = enc_value(other)
+        assert not _is_none_value(self)
+        assert not _is_none_value(other)
         return Expression(op, (other, self), inline=True)
 
     def __uop_expr__(self, op, *, params=None, inline=False):
         """unary expression"""
         assert isinstance(op, str)
+        assert not _is_none_value(self)
         return Expression(op, (self,), params=params, inline=inline, method=not inline)
 
     def __triop_expr__(self, op, x, y, inline=False, method=False):
@@ -187,6 +263,9 @@ class Term(PreTerm, ABC):
             x = enc_value(x)
         if not isinstance(y, Term):
             y = enc_value(y)
+        assert not _is_none_value(self)
+        assert not _is_none_value(x)
+        assert not _is_none_value(y)
         return Expression(op, (self, x, y), inline=inline, method=method)
 
     # try to get at == and other comparison operators
@@ -505,14 +584,15 @@ class Term(PreTerm, ABC):
         return self.__uop_expr__("is_bad")
 
     def if_else(self, x, y):
+        # could check if x and y are compatible types
         return self.__triop_expr__("if_else", x, y, method=True)
 
     def is_in(self, x):
-        return self.__op_expr__("is_in", x, inline=False, method=True)
+        return self.__op_expr__("is_in", x, inline=False, method=True, check_types=False,)
 
     def concat(self, x):
         # TODO: see if we can format back to infix notation
-        return self.__op_expr__("concat", x, inline=False, method=True)
+        return self.__op_expr__("concat", x, inline=False, method=True, check_types=False,)
 
     def coalesce(self, x):
         # TODO: see if we can format back to infix notation
@@ -547,26 +627,26 @@ class Term(PreTerm, ABC):
             format = Value("%Y-%m-%d %H:%M:%S")
         assert isinstance(format, Value)
         return self.__op_expr__(
-            "parse_datetime", other=format, inline=False, method=True
+            "parse_datetime", other=format, inline=False, method=True, check_types=False,
         )
 
     def parse_date(self, format=None):
         if format is None:
             format = Value("%Y-%m-%d")
-        return self.__op_expr__("parse_date", other=format, inline=False, method=True)
+        return self.__op_expr__("parse_date", other=format, inline=False, method=True, check_types=False,)
 
     def format_datetime(self, format=None):
         if format is None:
             format = Value("%Y-%m-%d %H:%M:%S")
         assert isinstance(format, Value)
         return self.__op_expr__(
-            "format_datetime", other=format, inline=False, method=True
+            "format_datetime", other=format, inline=False, method=True, check_types=False,
         )
 
     def format_date(self, format=None):
         if format is None:
             format = Value("%Y-%m-%d")
-        return self.__op_expr__("format_date", other=format, inline=False, method=True)
+        return self.__op_expr__("format_date", other=format, inline=False, method=True, check_types=False,)
 
     def dayofweek(self):
         return self.__uop_expr__("dayofweek")
@@ -593,7 +673,7 @@ class Term(PreTerm, ABC):
         return self.__op_expr__("timestamp_diff", other, inline=False, method=True)
 
     def date_diff(self, other):
-        return self.__op_expr__("date_diff", other, inline=False, method=True)
+        return self.__op_expr__("date_diff", other, inline=False, method=True, check_types=False,)
 
     # noinspection PyPep8Naming
     def base_Sunday(self):
@@ -610,7 +690,7 @@ def kop_expr(op, args, inline=False, method=False):
 class Value(Term):
     def __init__(self, value):
         allowed = {
-            data_algebra.util.map_type_to_canonical(t) for t in [int, float, str, bool]
+            data_algebra.util.map_type_to_canonical(t) for t in [int, float, str, bool, type(None)]
         }
         disaallowed = {data_algebra.util.map_type_to_canonical(type(value))} - allowed
         if len(disaallowed) != 0:
@@ -639,8 +719,8 @@ class Value(Term):
     def evaluate(self, data_frame):
         return self.value
 
-    def to_python(self, *, want_inline_parens=False):
-        return self.value.__repr__()
+    def to_python(self, *, want_inline_parens : bool = False) -> PythonText:
+        return PythonText(self.value.__repr__(), is_in_parens=False)
 
     # don't collect -5 as a complex expression
     def __neg__(self):
@@ -695,14 +775,14 @@ class ListTerm(PreTerm):
             res[i] = vi
         return res
 
-    def to_python(self, *, want_inline_parens=False):
+    def to_python(self, *, want_inline_parens : bool = False) -> PythonText:
         def li_to_python(value):
             try:
-                return value.to_python(want_inline_parens=want_inline_parens)
+                return str(value.to_python(want_inline_parens=False))
             except AttributeError:
                 return str(value)  # TODO: check if this should be repr?
 
-        return "[" + ", ".join([li_to_python(ai) for ai in self.value]) + "]"
+        return PythonText("[" + ", ".join([li_to_python(ai) for ai in self.value]) + "]", is_in_parens=False)
 
     def get_column_names(self, columns_seen):
         for ti in self.value:
@@ -757,8 +837,8 @@ class ColumnReference(Term):
     def replace_view(self, view):
         return ColumnReference(view=view, column_name=self.column_name)
 
-    def to_python(self, want_inline_parens=False):
-        return self.column_name
+    def to_python(self, *, want_inline_parens : bool = False) -> PythonText:
+        return PythonText(self.column_name, is_in_parens=False)
 
     def get_column_names(self, columns_seen):
         columns_seen.add(self.column_name)
@@ -907,32 +987,36 @@ class Expression(Term):
             pass
         raise KeyError(f"function {self.op} not found")
 
-    def to_python(self, *, want_inline_parens=False):
-        subs = [ai.to_python(want_inline_parens=True) for ai in self.args]
-        if len(subs) <= 0:
-            return self.op + "()"
-        if len(subs) == 1:
+    def to_python(self, *, want_inline_parens : bool = False) -> PythonText:
+        n_args = len(self.args)
+        if n_args <= 0:
+            return PythonText(self.op + "()", is_in_parens=False)
+        if n_args == 1:
+            sub_0 = self.args[0].to_python(want_inline_parens=False)
             if self.inline:
-                return self.op + self.args[0].to_python(want_inline_parens=True)
+                if sub_0.is_in_parens:
+                    return PythonText(self.op + str(sub_0), is_in_parens=False)
+                return PythonText(self.op + '(' + str(sub_0) + ')', is_in_parens=False)
             if self.method:
-                if isinstance(self.args[0], ColumnReference):
-                    return subs[0] + "." + self.op + "()"
-                else:
-                    return "(" + subs[0] + ")." + self.op + "()"
+                if sub_0.is_in_parens or isinstance(self.args[0], ColumnReference):
+                    return PythonText(str(sub_0) + "." + self.op + "()", is_in_parens=False)
+                return PythonText("(" + str(sub_0) + ")." + self.op + "()", is_in_parens=False)
         if self.inline:
-            result = ""
+            subs = [str(ai.to_python(want_inline_parens=True)) for ai in self.args]
+            result = (" " + self.op + " ").join(subs)
             if want_inline_parens:
-                result = result + "("
-            result = result + (" " + self.op + " ").join(subs)
-            if want_inline_parens:
-                result = result + ")"
-            return result
+                return PythonText('(' + result + ')', is_in_parens=True)
+            return PythonText(result, is_in_parens=False)
+        subs = [ai.to_python(want_inline_parens=False) for ai in self.args]
+        subs_0 = subs[0]
+        subs = [str(si) for si in subs]
         if self.method:
-            if isinstance(self.args[0], ColumnReference):
-                return subs[0] + "." + self.op + "(" + ", ".join(subs[1:]) + ")"
+            if subs_0.is_in_parens or isinstance(self.args[0], ColumnReference):
+                return PythonText(subs[0] + "." + self.op + "(" + ", ".join(subs[1:]) + ")", is_in_parens=False)
             else:
-                return "(" + subs[0] + ")." + self.op + "(" + ", ".join(subs[1:]) + ")"
-        return self.op + "(" + ", ".join(subs) + ")"
+                return PythonText("(" + subs[0] + ")." + self.op + "(" + ", ".join(subs[1:]) + ")", is_in_parens=False)
+        # treat as fn call
+        return PythonText(self.op + '(' + ", ".join(subs) + ')', is_in_parens=False)
 
 
 # define with def so function has usable __name__
