@@ -1,4 +1,5 @@
 from abc import ABC
+from typing import List, Optional, Tuple
 
 import data_algebra.OrderedSet
 
@@ -8,6 +9,30 @@ import data_algebra.OrderedSet
 # TODO: build a term object that carries the column use information
 
 
+class SQLWithList:
+    """
+    Carry an ordered sequence of SQL steps for use with a SQL WITH statement.
+    """
+    def __init__(self, *, last_step: 'NearSQL', previous_steps: List[Tuple[str, 'NearSQLContainer']]):
+        assert isinstance(last_step, NearSQL)
+        assert isinstance(previous_steps, List)
+        assert all([isinstance(vi, Tuple)
+                    and (len(vi) == 2)
+                    and isinstance(vi[0], str)
+                    and isinstance(vi[1], NearSQLContainer) for vi in previous_steps])
+        # check no table like entities in previous steps
+        assert not any([v.near_sql.is_table for k, v in previous_steps])
+        # check keying is unique
+        previous_keys = {k for k, v in previous_steps}
+        if len(previous_keys) != len(previous_steps):
+            assert len(previous_keys) == len(previous_steps)  # so we can set a breakpoint
+        # check we are not coming up on a duplication
+        if not last_step.is_table:
+            assert last_step.quoted_query_name not in previous_keys
+        self.last_step = last_step
+        self.previous_steps = previous_steps
+
+
 class NearSQL(ABC):
     """
     Represent SQL queries in a mostly string-form
@@ -32,7 +57,7 @@ class NearSQL(ABC):
             near_sql=self, columns=columns, force_sql=force_sql, constants=constants
         )
 
-    def to_sql(
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -40,14 +65,11 @@ class NearSQL(ABC):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
+    ) -> List[str]:
         raise NotImplementedError("base method called")
 
-    # return a list where last element is a NearSQL previous elements are (name, NearSQLContainer) pairs
-    def to_with_form(self):
-        sequence = list()
-        sequence.append(self)
-        return sequence
+    def to_with_form(self) -> SQLWithList:
+        raise NotImplementedError("base method called")
 
 
 class NearSQLContainer:
@@ -72,8 +94,8 @@ class NearSQLContainer:
         if constants is not None:
             self.constants = constants.copy()
 
-    def to_sql(self, db_model, sql_format_options=None):
-        return self.near_sql.to_sql(
+    def to_sql_str_list(self, db_model, sql_format_options=None) -> List[str]:
+        return self.near_sql.to_sql_str_list(
             columns=self.columns,
             force_sql=self.force_sql,
             constants=self.constants,
@@ -91,42 +113,51 @@ class NearSQLContainer:
 
     # sequence: a list where last element is a NearSQLContainer previous elements are (name, NearSQLContainer) pairs
     # stub the replacement common table expression in a NearSQLContainer
-    def to_with_form_stub(self):
-        stub = None
+    def to_with_form_stub(self) -> Tuple['NearSQLContainer', List[Tuple[str, 'NearSQLContainer']]]:
         if self.near_sql.is_table:
-            stub = self
-            sequence = list()
-        else:
-            sequence = self.near_sql.to_with_form()
-            endi = len(sequence) - 1
-            last_step = sequence[endi]
-            sequence[endi] = NearSQLContainer(
-                near_sql=last_step,
-                columns=self.columns,
+            return self, []
+        in_with_form = self.near_sql.to_with_form()
+        sequence = in_with_form.previous_steps
+        stub = in_with_form.last_step
+        if not stub.is_table:
+            # replace step with a reference
+            if stub.quoted_query_name not in {k for k, v in sequence}:
+                sequence.append((
+                    stub.quoted_query_name,
+                    NearSQLContainer(
+                        near_sql=stub,
+                        force_sql=self.force_sql,
+                    )
+                ))
+            stub = NearSQLContainer(
+                near_sql=NearSQLCommonTableExpression(
+                    quoted_query_name=stub.quoted_query_name
+                ),
                 force_sql=self.force_sql,
-                constants=self.constants,
             )
-            endi = len(sequence) - 1
-            last_step = sequence[endi]
-            stub = last_step
-            if not stub.near_sql.is_table:
-                stub = NearSQLContainer(
-                    near_sql=NearSQLCommonTableExpression(
-                        quoted_query_name=last_step.near_sql.quoted_query_name
-                    ),
-                    force_sql=self.force_sql,
-                )
-            sequence[endi] = (last_step.near_sql.quoted_query_name, last_step)
         return stub, sequence
 
 
-class NearSQLCommonTableExpression(NearSQL):
-    def __init__(self, *, quoted_query_name):
+class NearSQLNamedEntity(NearSQL):
+    def __init__(self, *, terms, quoted_query_name):
         NearSQL.__init__(
-            self, terms=None, quoted_query_name=quoted_query_name, is_table=True,
+            self, terms=terms, quoted_query_name=quoted_query_name, is_table=True,
         )
 
-    def to_sql(
+    def to_with_form(self) -> SQLWithList:
+        return SQLWithList(last_step=self, previous_steps=[])
+
+    def to_sql_str_list(self, *,
+                        columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+                        ) -> List[str]:
+        raise NotImplementedError("abstract class method called")
+
+
+class NearSQLCommonTableExpression(NearSQLNamedEntity):
+    def __init__(self, *, quoted_query_name):
+        NearSQLNamedEntity.__init__(self, terms=None, quoted_query_name=quoted_query_name)
+
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -134,8 +165,8 @@ class NearSQLCommonTableExpression(NearSQL):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
-        return db_model.nearsqlcte_to_sql_(
+    ) -> List[str]:
+        return db_model.nearsqlcte_to_sql_str_list_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
@@ -144,14 +175,12 @@ class NearSQLCommonTableExpression(NearSQL):
         )
 
 
-class NearSQLTable(NearSQL):
+class NearSQLTable(NearSQLNamedEntity):
     def __init__(self, *, terms, quoted_query_name, quoted_table_name):
-        NearSQL.__init__(
-            self, terms=terms, quoted_query_name=quoted_query_name, is_table=True,
-        )
+        NearSQLNamedEntity.__init__(self, terms=terms, quoted_query_name=quoted_query_name)
         self.quoted_table_name = quoted_table_name
 
-    def to_sql(
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -159,8 +188,8 @@ class NearSQLTable(NearSQL):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
-        return db_model.nearsqltable_to_sql_(
+    ) -> List[str]:
+        return db_model.nearsqltable_to_sql_str_list_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
@@ -184,7 +213,7 @@ class NearSQLUnaryStep(NearSQL):
         assert isinstance(mergeable, bool)
         assert isinstance(sub_sql, NearSQLContainer)
         assert isinstance(suffix, (list, type(None)))
-        if not suffix is None:
+        if suffix is not None:
             assert all([isinstance(v, str) for v in suffix])
         NearSQL.__init__(
             self,
@@ -202,7 +231,7 @@ class NearSQLUnaryStep(NearSQL):
         else:
             self.mergeable = False
 
-    def to_sql(
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -210,20 +239,18 @@ class NearSQLUnaryStep(NearSQL):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
-        return db_model.nearsqlunary_to_sql_(
+    ) -> List[str]:
+        return db_model.nearsqlunary_to_sql_str_list_(
             near_sql=self,
             columns=columns,
             constants=constants,
             sql_format_options=sql_format_options,
         )
 
-    def to_with_form(self):
+    def to_with_form(self) -> SQLWithList:
         if self.sub_sql.near_sql.is_table:
-            # tables don't need to be re-encoded
-            sequence = list()
-            sequence.append(self)
-            return sequence
+            # table references don't need to be re-encoded
+            return SQLWithList(last_step=self, previous_steps=[])
         # non-trivial sequence
         stub, sequence = self.sub_sql.to_with_form_stub()
         stubbed_step = NearSQLUnaryStep(
@@ -233,8 +260,7 @@ class NearSQLUnaryStep(NearSQL):
             suffix=self.suffix,
             annotation=self.annotation,
         )
-        sequence.append(stubbed_step)
-        return sequence
+        return SQLWithList(last_step=stubbed_step, previous_steps=sequence)
 
 
 class NearSQLBinaryStep(NearSQL):
@@ -252,7 +278,7 @@ class NearSQLBinaryStep(NearSQL):
         assert isinstance(sub_sql1, NearSQLContainer)
         assert isinstance(sub_sql2, NearSQLContainer)
         assert isinstance(suffix, (list, type(None)))
-        if not suffix is None:
+        if suffix is not None:
             assert all([isinstance(v, str) for v in suffix])
         assert isinstance(joiner, str)
         NearSQL.__init__(
@@ -266,7 +292,7 @@ class NearSQLBinaryStep(NearSQL):
         self.sub_sql2 = sub_sql2
         self.suffix = suffix
 
-    def to_sql(
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -274,8 +300,8 @@ class NearSQLBinaryStep(NearSQL):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
-        return db_model.nearsqlbinary_to_sql_(
+    ) -> List[str]:
+        return db_model.nearsqlbinary_to_sql_str_list_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
@@ -284,12 +310,10 @@ class NearSQLBinaryStep(NearSQL):
             quoted_query_name=self.quoted_query_name,
         )
 
-    def to_with_form(self):
+    def to_with_form(self) -> SQLWithList:
         if self.sub_sql1.near_sql.is_table and self.sub_sql2.near_sql.is_table:
-            # tables don't need to be re-encoded
-            sequence = list()
-            sequence.append(self)
-            return sequence
+            # tables references don't need to be re-encoded
+            return SQLWithList(last_step=self, previous_steps=[])
         # non-trivial sequence
         stub1, sequence1 = self.sub_sql1.to_with_form_stub()
         stub2, sequence2 = self.sub_sql2.to_with_form_stub()
@@ -302,7 +326,7 @@ class NearSQLBinaryStep(NearSQL):
         # assume any name collisions are the same table/common_table_expression
         for stepi in sequence2:
             nmi = stepi[0]
-            if not nmi in seen:
+            if nmi not in seen:
                 seen.add(nmi)
                 sequence.append(stepi)
         stubbed_step = NearSQLBinaryStep(
@@ -314,37 +338,39 @@ class NearSQLBinaryStep(NearSQL):
             suffix=self.suffix,
             annotation=self.annotation,
         )
-        sequence.append(stubbed_step)
-        return sequence
+        return SQLWithList(last_step=stubbed_step, previous_steps=sequence)
 
 
-class NearSQLq(NearSQL):
-    """
-    Adapter to wrap a pre-existing query as a NearSQL
-
-    """
-
+class NearSQLRawQStep(NearSQL):
     def __init__(
         self,
         *,
-        quoted_query_name,
-        query,
-        terms,
-        prev_quoted_query_name,
-        annotation=None,
+        prefix: List[str],
+        quoted_query_name: str,
+        sub_sql: NearSQLContainer,
+        suffix: Optional[List[str]] = None,
+        annotation: Optional[str] = None,
     ):
-        assert isinstance(query, str)
-        assert isinstance(prev_quoted_query_name, str)
+        assert isinstance(prefix, list)
+        assert len(prefix) > 0
+        assert all([isinstance(v, str) for v in prefix])
+        assert isinstance(quoted_query_name, str)
+        assert isinstance(sub_sql, NearSQLContainer)
+        assert isinstance(suffix, (list, type(None)))
+        if suffix is not None:
+            assert all([isinstance(v, str) for v in suffix])
+        assert isinstance(annotation, (str, type(None)))
         NearSQL.__init__(
             self,
-            terms=terms,
+            terms=None,  # not using this in this node
             quoted_query_name=quoted_query_name,
             annotation=annotation,
         )
-        self.query = query
-        self.prev_quoted_query_name = prev_quoted_query_name
+        self.prefix = prefix
+        self.sub_sql = sub_sql
+        self.suffix = suffix
 
-    def to_sql(
+    def to_sql_str_list(
         self,
         *,
         columns=None,
@@ -352,10 +378,25 @@ class NearSQLq(NearSQL):
         constants=None,
         db_model,
         sql_format_options=None,
-    ):
-        return db_model.nearsqlq_to_sql_(
+    ) -> List[str]:
+        return db_model.nearsqlrawq_to_sql_str_list_(
             near_sql=self,
             columns=columns,
             constants=constants,
             sql_format_options=sql_format_options,
         )
+
+    def to_with_form(self) -> SQLWithList:
+        if self.sub_sql.near_sql.is_table:
+            # table references don't need to be re-encoded
+            return SQLWithList(last_step=self, previous_steps=[])
+        # non-trivial sequence
+        stub, sequence = self.sub_sql.to_with_form_stub()
+        stubbed_step = NearSQLRawQStep(
+            prefix=self.prefix,
+            quoted_query_name=self.quoted_query_name,
+            sub_sql=stub,
+            suffix=self.suffix,
+            annotation=self.annotation,
+        )
+        return SQLWithList(last_step=stubbed_step, previous_steps=sequence)
