@@ -52,12 +52,16 @@ class OpC(data_algebra.data_ops_types.OperatorPlatform):
         :param allow_limited_tables: logical, if True allow execution on non-complete tables
         :return: table result
         """
+        assert not self.used_result
+        assert self.ops is not None
         self.used_result = True
-        return self.ops.ex(
+        res = self.ops.ex(
             data_model=data_model,
             narrow=narrow,
             allow_limited_tables=allow_limited_tables,
         )
+        self.ops = None
+        return res
 
     # noinspection PyPep8Naming
     def transform(self, X, *, data_model=None, narrow=True):
@@ -174,12 +178,11 @@ class OpC(data_algebra.data_ops_types.OperatorPlatform):
         return self
 
     def map_records(
-        self, blocks_in=None, blocks_out=None, strict=False
+        self, blocks_in=None, blocks_out=None
     ):
         self.ops.map_records(
             blocks_in=blocks_in,
             blocks_out=blocks_out,
-            strict=strict,
         )
         return self
 
