@@ -41,6 +41,7 @@ def formats_to_self(ops):
     str2 = repr(ops2)
     strings_match = str1 == str2  # probably too strict
     ops_match = ops == ops2
+    assert ops_match
     if strings_match and (not ops_match):
         raise Exception("strings match, but ops did not")
     pickle_string = pickle.dumps(ops)
@@ -233,7 +234,8 @@ def check_transform_on_handles(
             },
         ]
         for pm in partial_maps:
-            ops.eval(pm)
+            empty_res_i = ops.eval(pm)
+            assert set(empty_res_i.columns) == set(res.columns)
     # try any db paths
     if db_handles is not None:
         for db_handle in db_handles:
